"""The helper functions shared by all the CLI commands."""

# 1. Standard Python modules
from argparse import _HelpAction

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from .. import __version__
from ..resource import ResourceManager


def add_common_args(p):
    """Add the common arguments to the argument parser.

    Args:
        p (ArgumentParser): The parser to add arguments to.
    """
    # You must use the add_help=False argument to ArgumentParser or add_parser
    p.add_argument(
        '-h', '--help',
        action=_HelpAction,
        help="Show this help message and exit.",
    )
    p.add_argument(
        '-V', '--version',
        action='version',
        version='harmonica %s' % __version__,
        help="Show the harmonica version number and exit",
    )


def add_loc_model_args(p):
    """Add the location and model option arguments to the argument parser.

    Args:
        p (ArgumentParser): The parser to add arguments to.
    """
    # Required positional arguments
    # This combined argument is preferred but there is a bug in argparse with unpacking a multi-value metavar
    # p.add_argument(
    #     'loc',
    #     type=float,
    #     nargs=2,
    #     help='Desired location [latitude, longitude]',
    #     metavar=('LAT', 'LON'),
    # )
    p.add_argument(
        'lat',
        type=float,
        help='Desired latitude location',
        metavar='LAT',
    )
    p.add_argument(
        'lon',
        type=float,
        help='Desired longitude location',
        metavar='LON',
    )
    # Optional arguments
    p.add_argument(
        '-M', '--model',
        choices=ResourceManager.RESOURCES.keys(),
        default=ResourceManager.DEFAULT_RESOURCE,
        help='Optional constituent model specification, default: {}'.format(ResourceManager.DEFAULT_RESOURCE),
    )


def add_const_out_args(p):
    """Add the constituent output option arguments to the argument parser.

    Args:
        p (ArgumentParser): The parser to add arguments to.
    """
    p.add_argument(
        '-C', '--cons',
        nargs='+',
        default=None,
        help='Optional list of constituents to retrieve; retrieves all by default',
    )
    p.add_argument(
        '-P', '--positive_phase',
        action='store_true',
        default=False,
        help='Boolean to indicate whether phase should be all positive [0,360] (True) or [-180,180] (False; default)',
    )
    p.add_argument(
        '-O', '--output',
        default=None,
        help='Write output to specified file',
    )



"""The constituents CLI command."""

# 1. Standard Python modules
import argparse
import sys

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from .common import add_common_args, add_const_out_args, add_loc_model_args
from ..tidal_constituents import Constituents


DESCR = 'Get specified tidal constituents at specified locations.'
EXAMPLE = """
Example:

    harmonica constituents 38.375789 -74.943915 -C M2 K1 -M tpxo8
"""


def config_parser(p, sub=False):
    """Configure the command line arguments passed the constituents CLI command.

    Args:
        p (ArgumentParser): The argument parser
        sub (Optional[bool]): True if this is a resources subparser
    """
    # Subparser info
    if sub:
        p = p.add_parser(
            'constituents',
            description=DESCR,
            help=DESCR,
            epilog=EXAMPLE,
            add_help=False,
        )

    add_common_args(p)
    add_loc_model_args(p)
    add_const_out_args(p)


def parse_args(args):
    """Parse the command line arguments passed the constituents CLI command.

    Args:
        args (...): Variable length positional arguments

    Returns:
        ArgumentParser: The command line argument parser
    """
    p = argparse.ArgumentParser(
        description=DESCR,
        epilog=EXAMPLE,
        add_help=False,
    )
    config_parser(p)
    return p.parse_args(args)


def execute(args):
    """Execute the constituents CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    cons = Constituents(model=args.model).get_components(
        [(args.lat, args.lon)], cons=args.cons, positive_ph=args.positive_phase
    )
    out = cons.data[0].to_csv(args.output, sep='\t', header=True, index=True, index_label='constituent')
    if args.output is None:
        print(out)
    print("\nComplete.\n")


def main(args=None):
    """Entry point for the constituents CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    if not args:
        args = sys.argv[1:]
    try:
        execute(parse_args(args))
    except RuntimeError as e:
        print(str(e))
        sys.exit(1)
    return



"""Base class for the tidal database models."""

# 1. Standard Python modules
from abc import ABCMeta, abstractmethod
import math

# 2. Third party modules
import numpy
import pandas as pd

# 3. Aquaveo modules

# 4. Local modules
from .resource import ResourceManager


NCNST = 37

# Dictionary of NOAA constituent speed constants (deg/hr)
# Source: https://tidesandcurrents.noaa.gov
# The speed is the rate change in the phase of a constituent, and is equal to 360 degrees divided by the
# constituent period expressed in hours
# name: (speed, amplitude, frequency, ETRF)
NOAA_SPEEDS = {
    'OO1': (16.139101, 0.0, 0.00007824457305, 0.069),
    '2Q1': (12.854286, 0.0, 0.000062319338107, 0.069),
    '2MK3': (42.92714, 0.0, 0.000208116646659, 0.069),
    '2N2': (27.895355, 0.0, 0.000135240496464, 0.069),
    '2SM2': (31.015896, 0.0, 0.000150369306157, 0.069),
    'K1': (15.041069, 0.141565, 0.000072921158358, 0.736),
    'K2': (30.082138, 0.030704, 0.000145842317201, 0.693),
    'J1': (15.5854435, 0.0, 0.00007556036138, 0.069),
    'L2': (29.528479, 0.0, 0.000143158105531, 0.069),
    'LAM2': (29.455626, 0.0, 0.000142804901311, 0.069),
    'M1': (14.496694, 0.0, 0.000070281955336, 0.069),
    'M2': (28.984104, 0.242334, 0.000140518902509, 0.693),
    'M3': (43.47616, 0.0, 0.000210778353763, 0.069),
    'M4': (57.96821, 0.0, 0.000281037805017, 0.069),
    'M6': (86.95232, 0.0, 0.000421556708011, 0.069),
    'M8': (115.93642, 0.0, 0.000562075610519, 0.069),
    'MF': (1.0980331, 0.0, 0.000005323414692, 0.069),
    'MK3': (44.025173, 0.0, 0.000213440061351, 0.069),
    'MM': (0.5443747, 0.0, 0.000002639203022, 0.069),
    'MN4': (57.423832, 0.0, 0.000278398601995, 0.069),
    'MS4': (58.984104, 0.0, 0.000285963006842, 0.069),
    'MSF': (1.0158958, 0.0, 0.000004925201824, 0.069),
    'MU2': (27.968208, 0.0, 0.000135593700684, 0.069),
    'N2': (28.43973, 0.046398, 0.000137879699487, 0.693),
    'NU2': (28.512583, 0.0, 0.000138232903707, 0.069),
    'O1': (13.943035, 0.100514, 0.000067597744151, 0.695),
    'P1': (14.958931, 0.046834, 0.000072522945975, 0.706),
    'Q1': (13.398661, 0.019256, 0.000064958541129, 0.695),
    'R2': (30.041067, 0.0, 0.000145643201313, 0.069),
    'RHO': (13.471515, 0.0, 0.000065311745349, 0.069),
    'S1': (15.0, 0.0, 0.000072722052166, 0.069),
    'S2': (30.0, 0.112841, 0.000145444104333, 0.693),
    'S4': (60.0, 0.0, 0.000290888208666, 0.069),
    'S6': (90.0, 0.0, 0.000436332312999, 0.069),
    'SA': (0.0410686, 0.0, 0.000000199106191, 0.069),
    'SSA': (0.0821373, 0.0, 0.000000398212868, 0.069),
    'T2': (29.958933, 0.0, 0.000145245007353, 0.069),
}


def get_complex_components(amps, phases):
    """Get the real and imaginary components of amplitudes and phases.

    Args:
        amps (:obj:`list` of :obj:`float`): List of constituent amplitudes
        phases (:obj:`list` of :obj:`float`): List of constituent phases in radians

    Returns:
        :obj:`list` of :obj:`tuple` of :obj:`float`: The list of the complex components,
            e.g. [[real1, imag1], [real2, imag2]]

    """
    components = [[0.0, 0.0] for _ in range(len(amps))]
    for idx, (amp, phase) in enumerate(zip(amps, phases)):
        components[idx][0] = amp * math.cos(phase)
        components[idx][1] = amp * math.sin(phase)
    return components


def convert_coords(coords, zero_to_360=False):
    """Convert latitude coordinates to [-180, 180] or [0, 360].

    Args:
        coords (:obj:`list` of :obj:`tuple` of :obj:`float`): latitude [-90, 90] and longitude [-180 180] or [0 360]
            of the requested point.
        zero_to_360 (:obj:`bool`, optional) If True, coordinates will be converted to the [0, 360] range. If False,
            coordinates will be converted to the [-180, 180] range.

    Returns:
        :obj:`list` of :obj:`tuple` of :obj:`float`: The list of converted coordinates. None if a coordinate out of
            range was encountered

    """
    # Make sure point locations are valid lat/lon
    for idx, pt in enumerate(coords):
        y_lat = pt[0]
        x_lon = pt[1]
        if x_lon < 0.0:
            x_lon += 360.0
        if not zero_to_360 and x_lon > 180.0:
            x_lon -= 360.0

        bad_lat = y_lat > 90.0 or y_lat < -90.0  # Invalid latitude
        bad_lon = not zero_to_360 and (x_lon > 180.0 or x_lon < -180.0)  # Invalid [-180, 180]
        bad_lon = bad_lon or (zero_to_360 and (x_lon > 360.0 or x_lon < 0.0))  # Invalid [0, 360]
        if bad_lat or bad_lon:
            # ERROR: Not in latitude/longitude
            return None
        else:
            coords[idx] = (y_lat, x_lon)
    return coords


class OrbitVariables(object):
    """Container for variables used in astronomical equations.

    Attributes:
        astro (:obj:`pytides.astro.astro`): Orbit variables obtained from pytides.
        grterm (:obj:`dict` of :obj:`float`): Dictionary of equilibrium arguments where the key is constituent name
        nodfac (:obj:`dict` of :obj:`float`): Dictionary of nodal factors where the key is constituent name

    """
    def __init__(self):
        """Construct the container."""
        self.astro = {}
        self.grterm = {con: 0.0 for con in NOAA_SPEEDS}
        self.nodfac = {con: 0.0 for con in NOAA_SPEEDS}


class TidalDB(object):
    """The base class for extracting tidal data from a database.

    Attributes:
        orbit (:obj:`OrbitVariables`): The orbit variables.
        data (:obj:`list` of :obj:`pandas.DataFrame`): List of the constituent component DataFrames with one
            per point location requested from get_components(). Intended return value of get_components().
        resources (:obj:`harmonica.resource.ResourceManager`): Manages fetching of tidal data

    """
    """(:obj:`list` of :obj:`float`): The starting days of the months (non-leap year)."""
    day_t = [0.0, 31.0, 59.0, 90.0, 120.0, 151.0, 181.0, 212.0, 243.0, 273.0, 304.0, 334.0]

    def __init__(self, model):
        """Base class constructor for the tidal extractors.

        Args:
            model (str): The name of the model. See resource.py for supported models.
        """
        self.orbit = OrbitVariables()
        # constituent information dataframe:
        #   amplitude (meters)
        #   phase (degrees)
        #   speed (degrees/hour, UTC/GMT)
        self.data = []
        self.model = model
        self.resources = ResourceManager(self.model)

    __metaclass__ = ABCMeta

    @abstractmethod
    def get_components(self, locs, cons, positive_ph):
        """Abstract method to get amplitude, phase, and speed of specified constituents at specified point locations.

        Args:
            locs (:obj:`list` of :obj:`tuple` of :obj:`float`): latitude [-90, 90] and longitude [-180 180] or [0 360]
                of the requested points.
            cons (:obj:`list` of :obj:`str`, optional): List of the constituent names to get amplitude and phase for. If
                not supplied, all valid constituents will be extracted.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).

        Returns:
           :obj:`list` of :obj:`pandas.DataFrame`: Implementations should return a list of dataframes of constituent
                information including amplitude (meters), phase (degrees) and speed (degrees/hour, UTC/GMT). The list is
                parallel with locs, where each element in the return list is the constituent data for the corresponding
                element in locs. Empty list on error. Note that function uses fluent interface pattern.
        """
        pass

    def have_constituent(self, name):
        """Determine if a constituent is valid for this tidal extractor.

        Args:
            name (str): The name of the constituent.

        Returns:
            bool: True if the constituent is valid, False otherwise
        """
        return name.upper() in self.resources.available_constituents()

    def get_nodal_factor(self, names, timestamp, timestamp_middle):
        """Get the nodal factor for specified constituents at a specified time.

        Args:
            names (:obj:`list` of :obj:`str`): Names of the constituents to get nodal factors for
            timestamp (datetime.datetime): Start date and time to extract constituent arguments at
            timestamp_middle (datetime.datetime): Date and time to consider as the middle of the series.

        Returns:
            :obj:`pandas.DataFrame`: Constituent data frames. Each row contains frequency, earth tidal reduction factor,
                amplitude, nodal factor, and equilibrium argument for one of the specified constituents. Rows labeled by
                constituent name.
        """
        con_data = pd.DataFrame(columns=['amplitude', 'frequency', 'speed', 'earth_tide_reduction_factor',
                                         'equilibrium_argument', 'nodal_factor'])
        self.get_eq_args(timestamp, timestamp_middle)
        for name in names:
            name = name.upper()
            if name not in NOAA_SPEEDS:
                con_data.loc[name] = [numpy.nan, numpy.nan, numpy.nan, numpy.nan, numpy.nan, numpy.nan]
            else:
                equilibrium_arg = 0.0
                nodal_factor = self.orbit.nodfac[name]
                if nodal_factor != 0.0:
                    equilibrium_arg = self.orbit.grterm[name]
                con_data.loc[name] = [
                    NOAA_SPEEDS[name][1], NOAA_SPEEDS[name][2], NOAA_SPEEDS[name][0], NOAA_SPEEDS[name][3],
                    equilibrium_arg, nodal_factor
                ]
        return con_data

    def get_eq_args(self, timestamp, timestamp_middle):
        """Get equilibrium arguments at a starting time.

        Args:
            timestamp (datetime.datetime): Date and time to extract constituent arguments at
            timestamp_middle (datetime.datetime): Date and time to consider as the middle of the series
        """
        self.nfacs(timestamp_middle)
        self.gterms(timestamp, timestamp_middle)

    @staticmethod
    def angle(a_number):
        """Converts the angle to be within 0-360.

        Args:
            a_number (float): The angle to convert.

        Returns:
            The angle converted to be within 0-360.
        """
        return a_number % 360.0

    def set_orbit(self, timestamp):
        """Determination of primary and secondary orbital functions.

        Args:
           timestamp (datetime.datetime): Date and time to extract constituent arguments at.
        """
        # We used to rely on pytides for astronomical computations, but it was giving was different results from
        # the tide_fac Fortran utility.
        # from pytides.astro import astro
        # self.orbit.astro = astro(timestamp)

        # Ported code from tide_fac.f
        dayj = timestamp.timetuple().tm_yday  # ordinal day number
        # pi180 = 3.14159265 / 180.0
        x = int((timestamp.year - 1901.) / 4.0)
        dyr = timestamp.year - 1900.0
        dday = dayj + x - 1.0
        # DN IS THE MOON'S NODE (CAPITAL N, TABLE 1, SCHUREMAN)
        dn = 259.1560564 - 19.328185764 * dyr - 0.0529539336 * dday - 0.0022064139 * timestamp.hour
        dn = self.angle(dn)
        n = math.radians(dn)
        # DP IS THE LUNAR PERIGEE (SMALL P, TABLE 1)
        dp = 334.3837214 + 40.66246584 * dyr + 0.111404016 * dday + 0.004641834 * timestamp.hour
        dp = self.angle(dp)
        i = math.acos(0.9136949 - 0.0356926 * math.cos(n))
        di = self.angle(math.degrees(i))
        nu = math.asin(0.0897056 * math.sin(n) / math.sin(i))
        dnu = math.degrees(nu)
        xi = n - 2.0 * math.atan(0.64412 * math.tan(n / 2.0)) - nu
        dxi = math.degrees(xi)
        dpc = self.angle(dp - dxi)
        # DH IS THE MEAN LONGITUDE OF THE SUN (SMALL H, TABLE 1)
        dh = 280.1895014 - 0.238724988 * dyr + 0.9856473288 * dday + 0.0410686387 * timestamp.hour
        dh = self.angle(dh)
        # DP1 IS THE SOLAR PERIGEE (SMALL P1, TABLE 1)
        dp1 = 281.2208569 + 0.01717836 * dyr + 0.000047064 * dday + 0.000001961 * timestamp.hour
        dp1 = self.angle(dp1)
        # DS IS THE MEAN LONGITUDE OF THE MOON (SMALL S, TABLE 1)
        ds = 277.0256206 + 129.38482032 * dyr + 13.176396768 * dday + 0.549016532 * timestamp.hour
        ds = self.angle(ds)
        nup = math.atan(math.sin(nu) / (math.cos(nu) + 0.334766 / math.sin(2.0 * i)))
        dnup = math.degrees(nup)
        nup2 = math.atan(math.sin(2.0 * nu) / (math.cos(2.0 * nu) + 0.0726184 / math.sin(i) ** 2)) / 2.0
        dnup2 = math.degrees(nup2)

        self.orbit.astro = {
            'ds': ds,
            'dp': dp,
            'dh': dh,
            'dp1': dp1,
            'dn': dn,
            'di': di,
            'dnu': dnu,
            'dxi': dxi,
            'dnup': dnup,
            'dnup2': dnup2,
            'dpc': dpc,
        }

    def nfacs(self, timestamp):
        """Calculates node factors for constituent tidal signal.

        Args:
            timestamp (datetime.datetime): Date and time to extract constituent arguments at

        Returns:
            The same values as found in table 14 of Schureman.
        """
        # Ported code from tide_fac.f
        self.set_orbit(timestamp)
        # n = math.radians(self.orbit.astro['dn'])
        i = math.radians(self.orbit.astro['di'])
        nu = math.radians(self.orbit.astro['dnu'])
        # xi = math.radians(self.orbit.astro['dxi'])
        # p = math.radians(self.orbit.astro['dp'])
        # pc = math.radians(self.orbit.astro['dpc'])
        sini = math.sin(i)
        sini2 = math.sin(i / 2.0)
        sin2i = math.sin(2.0 * i)
        cosi2 = math.cos(i / 2.0)
        # tani2 = math.tan(i / 2.0)
        # EQUATION 197, SCHUREMAN
        # qainv = math.sqrt(2.310 + 1.435 * math.cos(2.0 * pc))
        # EQUATION 213, SCHUREMAN
        # rainv = math.sqrt(1.0 - 12.0 * tani2 ** 2 * math.cos(2.0 * pc) + 36.0 * tani2 ** 4)
        # VARIABLE NAMES REFER TO EQUATION NUMBERS IN SCHUREMAN
        eq73 = (2.0 / 3.0 - sini ** 2) / 0.5021
        eq74 = sini ** 2 / 0.1578
        eq75 = sini * cosi2 ** 2 / 0.37988
        eq76 = math.sin(2 * i) / 0.7214
        eq77 = sini * sini2 ** 2 / 0.0164
        eq78 = cosi2 ** 4 / 0.91544
        eq149 = cosi2 ** 6 / 0.8758
        # eq207 = eq75 * qainv
        # eq215 = eq78 * rainv
        eq227 = math.sqrt(0.8965 * sin2i ** 2 + 0.6001 * sin2i * math.cos(nu) + 0.1006)
        eq235 = 0.001 + math.sqrt(19.0444 * sini ** 4 + 2.7702 * sini ** 2 * math.cos(2.0 * nu) + 0.0981)
        # NODE FACTORS FOR 37 CONSTITUENTS:
        self.orbit.nodfac['M2'] = eq78
        self.orbit.nodfac['S2'] = 1.0
        self.orbit.nodfac['N2'] = eq78
        self.orbit.nodfac['K1'] = eq227
        self.orbit.nodfac['M4'] = self.orbit.nodfac['M2'] ** 2
        self.orbit.nodfac['O1'] = eq75
        self.orbit.nodfac['M6'] = self.orbit.nodfac['M2'] ** 3
        self.orbit.nodfac['MK3'] = self.orbit.nodfac['M2'] * self.orbit.nodfac['K1']
        self.orbit.nodfac['S4'] = 1.0
        self.orbit.nodfac['MN4'] = self.orbit.nodfac['M2'] ** 2
        self.orbit.nodfac['NU2'] = eq78
        self.orbit.nodfac['S6'] = 1.0
        self.orbit.nodfac['MU2'] = eq78
        self.orbit.nodfac['2N2'] = eq78
        self.orbit.nodfac['OO1'] = eq77
        self.orbit.nodfac['LAM2'] = eq78
        self.orbit.nodfac['S1'] = 1.0
        # EQUATION 207 NOT PRODUCING CORRECT ANSWER FOR M1
        # SET NODE FACTOR FOR M1 = 0 UNTIL CAN FURTHER RESEARCH
        self.orbit.nodfac['M1'] = 0.0  # = eq207
        self.orbit.nodfac['J1'] = eq76
        self.orbit.nodfac['MM'] = eq73
        self.orbit.nodfac['SSA'] = 1.0
        self.orbit.nodfac['SA'] = 1.0
        self.orbit.nodfac['MSF'] = eq78
        self.orbit.nodfac['MF'] = eq74
        self.orbit.nodfac['RHO'] = eq75
        self.orbit.nodfac['Q1'] = eq75
        self.orbit.nodfac['T2'] = 1.0
        self.orbit.nodfac['R2'] = 1.0
        self.orbit.nodfac['2Q1'] = eq75
        self.orbit.nodfac['P1'] = 1.0
        self.orbit.nodfac['2SM2'] = eq78
        self.orbit.nodfac['M3'] = eq149
        # EQUATION 215 NOT PRODUCING CORRECT ANSWER FOR L2
        # SET NODE FACTOR FOR L2 = 0 UNTIL CAN FURTHER RESEARCH
        self.orbit.nodfac['L2'] = 0.0  # = eq215
        self.orbit.nodfac['2MK3'] = self.orbit.nodfac['M2'] ** 2 * self.orbit.nodfac['K1']
        self.orbit.nodfac['K2'] = eq235
        self.orbit.nodfac['M8'] = self.orbit.nodfac['M2'] ** 4
        self.orbit.nodfac['MS4'] = eq78

    def gterms(self, timestamp, timestamp_middle):
        """Determines the Greenwich equilibrium terms.

        Args:
            timestamp (datetime.datetime): Start date and time to extract constituent arguments at
            timestamp_middle (datetime.datetime): Date and time to consider as the middle of the series
        Returns:
            The same values as found in table 15 of Schureman.
        """
        # Ported code from tide_fac.f
        # OBTAINING ORBITAL VALUES AT BEGINNING OF SERIES FOR V0
        self.set_orbit(timestamp)
        s = self.orbit.astro['ds']
        p = self.orbit.astro['dp']
        h = self.orbit.astro['dh']
        p1 = self.orbit.astro['dp1']
        t = self.angle(180.0 + timestamp.hour * (360.0 / 24.0))

        # OBTAINING ORBITAL VALUES AT MIDDLE OF SERIES FOR U
        self.set_orbit(timestamp_middle)
        nu = self.orbit.astro['dnu']
        xi = self.orbit.astro['dxi']
        nup = self.orbit.astro['dnup']
        nup2 = self.orbit.astro['dnup2']

        # SUMMING TERMS TO OBTAIN EQUILIBRIUM ARGUMENTS
        self.orbit.grterm['M2'] = 2.0 * (t - s + h) + 2.0 * (xi - nu)
        self.orbit.grterm['S2'] = 2.0 * t
        self.orbit.grterm['N2'] = 2.0 * (t + h) - 3.0 * s + p + 2.0 * (xi - nu)
        self.orbit.grterm['K1'] = t + h - 90.0 - nup
        self.orbit.grterm['M4'] = 4.0 * (t - s + h) + 4.0 * (xi - nu)
        self.orbit.grterm['O1'] = t - 2.0 * s + h + 90.0 + 2.0 * xi - nu
        self.orbit.grterm['M6'] = 6.0 * (t - s + h) + 6.0 * (xi - nu)
        self.orbit.grterm['MK3'] = 3.0 * (t + h) - 2.0 * s - 90.0 + 2.0 * (xi - nu) - nup
        self.orbit.grterm['S4'] = 4.0 * t
        self.orbit.grterm['MN4'] = 4.0 * (t + h) - 5.0 * s + p + 4.0 * (xi - nu)
        self.orbit.grterm['NU2'] = 2.0 * t - 3.0 * s + 4.0 * h - p + 2.0 * (xi - nu)
        self.orbit.grterm['S6'] = 6.0 * t
        self.orbit.grterm['MU2'] = 2.0 * (t + 2.0 * (h - s)) + 2.0 * (xi - nu)
        self.orbit.grterm['2N2'] = 2.0 * (t - 2.0 * s + h + p) + 2.0 * (xi - nu)
        self.orbit.grterm['OO1'] = t + 2.0 * s + h - 90.0 - 2.0 * xi - nu
        self.orbit.grterm['LAM2'] = 2.0 * t - s + p + 180.0 + 2.0 * (xi - nu)
        self.orbit.grterm['S1'] = t
        i = math.radians(self.orbit.astro['di'])
        pc = math.radians(self.orbit.astro['dpc'])
        top = (5.0 * math.cos(i) - 1.0) * math.sin(pc)
        bottom = (7.0 * math.cos(i) + 1.0) * math.cos(pc)
        q = self.angle(math.degrees(math.atan2(top, bottom)))
        self.orbit.grterm['M1'] = t - s + h - 90.0 + xi - nu + q
        self.orbit.grterm['J1'] = t + s + h - p - 90.0 - nu
        self.orbit.grterm['MM'] = s - p
        self.orbit.grterm['SSA'] = 2.0 * h
        self.orbit.grterm['SA'] = h
        self.orbit.grterm['MSF'] = 2.0 * (s - h)
        self.orbit.grterm['MF'] = 2.0 * s - 2.0 * xi
        self.orbit.grterm['RHO'] = t + 3.0 * (h - s) - p + 90.0 + 2.0 * xi - nu
        self.orbit.grterm['Q1'] = t - 3.0 * s + h + p + 90.0 + 2.0 * xi - nu
        self.orbit.grterm['T2'] = 2.0 * t - h + p1
        self.orbit.grterm['R2'] = 2.0 * t + h - p1 + 180.0
        self.orbit.grterm['2Q1'] = t - 4.0 * s + h + 2.0 * p + 90.0 + 2.0 * xi - nu
        self.orbit.grterm['P1'] = t - h + 90.0
        self.orbit.grterm['2SM2'] = 2.0 * (t + s - h) + 2.0 * (nu - xi)
        self.orbit.grterm['M3'] = 3.0 * (t - s + h) + 3.0 * (xi - nu)
        r = math.sin(2.0 * pc) / ((1.0 / 6.0) * (1.0 / math.tan(0.5 * i)) ** 2 - math.cos(2.0 * pc))
        r = math.degrees(math.atan(r))
        self.orbit.grterm['L2'] = 2.0 * (t + h) - s - p + 180.0 + 2.0 * (xi - nu) - r
        self.orbit.grterm['2MK3'] = 3.0 * (t + h) - 4.0 * s + 90.0 + 4.0 * (xi - nu) + nup
        self.orbit.grterm['K2'] = 2.0 * (t + h) - 2.0 * nup2
        self.orbit.grterm['M8'] = 8.0 * (t - s + h) + 8.0 * (xi - nu)
        self.orbit.grterm['MS4'] = 2.0 * (2.0 * t - s + h) + 2.0 * (xi - nu)
        for con, value in self.orbit.grterm.items():
            self.orbit.grterm[con] = self.angle(value)



"""Entry point for the CLI tools."""

# 1. Standard Python modules

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from harmonica.cli import main

if __name__ == "__main__":
    main.main()



"""This module contains the tidal database extractor for the LeProvost tidal database."""

# 1. Standard Python modules
import math
import os

# 2. Third party modules
import numpy
import pandas as pd

# 3. Aquaveo modules

# 4. Local modules
from .resource import ResourceManager
from .tidal_database import convert_coords, get_complex_components, NOAA_SPEEDS, TidalDB


DEFAULT_LEPROVOST_RESOURCE = 'leprovost'


class LeProvostDB(TidalDB):
    """Extractor class for the LeProvost tidal database.

    """
    def __init__(self, model=DEFAULT_LEPROVOST_RESOURCE):
        """Constructor for the LeProvost tidal database extractor.

        Args:
            model (:obj:`str`, optional): Name of the LeProvost tidal database version. Defaults to the freely
                distributed but outdated 'leprovost' version. See resource.py for additional supported models.
        """
        model = model.lower()  # Be case-insensitive
        if model not in ResourceManager.LEPROVOST_MODELS:  # Check for valid LeProvost model
            raise ValueError("\'{}\' is not a supported LeProvost model. Must be one of: {}.".format(
                model, ", ".join(ResourceManager.LEPROVOST_MODELS).strip()
            ))
        super().__init__(model)

    def get_components(self, locs, cons=None, positive_ph=False):
        """Get the amplitude, phase, and speed of specified constituents at specified point locations.

        Args:
            locs (:obj:`list` of :obj:`tuple` of :obj:`float`): latitude [-90, 90] and longitude [-180 180] or [0 360]
                of the requested points.
            cons (:obj:`list` of :obj:`str`, optional): List of the constituent names to get amplitude and phase for. If
                not supplied, all valid constituents will be extracted.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).

        Returns:
           :obj:`list` of :obj:`pandas.DataFrame`: A list of dataframes of constituent information including
                amplitude (meters), phase (degrees) and speed (degrees/hour, UTC/GMT). The list is parallel with locs,
                where each element in the return list is the constituent data for the corresponding element in locs.
                Empty list on error. Note that function uses fluent interface pattern.
        """
        # If no constituents specified, extract all valid constituents.
        if not cons:
            cons = list(self.resources.available_constituents())
        else:  # Be case-insensitive
            cons = [con.upper() for con in cons]

        # Make sure point locations are valid lat/lon
        locs = convert_coords(locs, self.model == "fes2014")
        if not locs:
            return self  # ERROR: Not in latitude/longitude

        # Accumulate each point's constituent rows and build one DataFrame per point at the end, instead of
        # writing DataFrame cells one at a time in the inner loop.
        rows = [{} for _ in range(len(locs))]
        dataset_atts = self.resources.model_atts.dataset_attributes()

        n_lat = dataset_atts['num_lats']
        n_lon = dataset_atts['num_lons']
        lat_min = -90.0
        lon_min = dataset_atts['min_lon']
        d_lat = 180.0 / (n_lat - 1)
        d_lon = 360.0 / n_lon

        filenames = []
        for dset_idx, dset in enumerate(self.resources.get_datasets(cons, filenames)):
            if self.model == 'leprovost':  # All constituents in one file with constituent name dataset.
                nc_names = [x.strip().upper() for x in dset[0].spectrum.data.tolist()]
            else:  # FES2014 has separate files for each constituent with no constituent name dataset.
                con_filenames = filenames[dset_idx]
                nc_names = [os.path.splitext(os.path.basename(filename))[0].upper() for filename in con_filenames]
            for con in sorted(set(cons) & set(nc_names)):
                # Select the constituent's grid(s) once per constituent and materialize to numpy so the point loop
                # indexes numpy directly instead of re-selecting and reading through xarray for every point.
                con_idx = nc_names.index(con)
                if self.model == 'leprovost':  # All constituents in one file.
                    amp_dset = dset[0].amplitude[con_idx].values
                    phase_dset = dset[0].phase[con_idx].values
                else:  # FES2014 has a separate file per constituent.
                    amp_dset = dset[con_idx].amplitude.values
                    phase_dset = dset[con_idx].phase.values
                for i, pt in enumerate(locs):
                    y_lat, x_lon = pt  # lat,lon not x,y
                    xlo = int((x_lon - lon_min) / d_lon) + 1
                    xlonlo = lon_min + (xlo - 1) * d_lon
                    xhi = xlo + 1
                    if xlo == n_lon:
                        xhi = 1
                    ylo = int((y_lat - lat_min) / d_lon) + 1
                    ylatlo = lat_min + (ylo - 1) * d_lat
                    yhi = ylo + 1
                    xlo -= 1
                    xhi -= 1
                    ylo -= 1
                    yhi -= 1
                    skip = False

                    # Make sure lat/lon coordinate is in the domain.
                    out_of_bounds = xlo > n_lon or xhi > n_lon or yhi > n_lat or ylo > n_lat
                    out_of_bounds = out_of_bounds or xlo < 0 or xhi < 0 or yhi < 0 or ylo < 0
                    if out_of_bounds:
                        skip = True
                    else:  # Make sure we have at least one neighbor with an active amplitude value.
                        # Read potential contributing amplitudes from the file.
                        xlo_yhi_amp = amp_dset[yhi][xlo]
                        xlo_ylo_amp = amp_dset[ylo][xlo]
                        xhi_yhi_amp = amp_dset[yhi][xhi]
                        xhi_ylo_amp = amp_dset[ylo][xhi]
                        hi_nan = numpy.isnan(xlo_yhi_amp) and numpy.isnan(xhi_yhi_amp)
                        lo_nan = numpy.isnan(xlo_ylo_amp) and numpy.isnan(xhi_ylo_amp)
                        if hi_nan and lo_nan:
                            skip = True
                        else:  # Make sure we have at least one neighbor with an active phase value.
                            # Read potential contributing phases from the file.
                            xlo_yhi_phase = math.radians(phase_dset[yhi][xlo])
                            xlo_ylo_phase = math.radians(phase_dset[ylo][xlo])
                            xhi_yhi_phase = math.radians(phase_dset[yhi][xhi])
                            xhi_ylo_phase = math.radians(phase_dset[ylo][xhi])
                            hi_nan = numpy.isnan(xlo_yhi_phase) and numpy.isnan(xhi_yhi_phase)
                            lo_nan = numpy.isnan(xlo_ylo_phase) and numpy.isnan(xhi_ylo_phase)
                            if hi_nan and lo_nan:
                                skip = True

                    if skip:
                        rows[i][con] = (numpy.nan, numpy.nan, numpy.nan)
                    else:
                        xratio = (x_lon - xlonlo) / d_lon
                        yratio = (y_lat - ylatlo) / d_lat

                        # Get the real and imaginary components from the amplitude and phases in the file. It
                        # would be better if these values were stored in the file like TPXO.
                        complex_comps = get_complex_components(
                            amps=[xlo_yhi_amp, xhi_yhi_amp, xlo_ylo_amp, xhi_ylo_amp],
                            phases=[xlo_yhi_phase, xhi_yhi_phase, xlo_ylo_phase, xhi_ylo_phase],
                        )

                        # Perform bi-linear interpolation from the four cell corners to the target point.
                        xcos = 0.0
                        xsin = 0.0
                        denom = 0.0
                        if not numpy.isnan(xlo_yhi_amp) and not numpy.isnan(xlo_yhi_phase):
                            xcos = xcos + complex_comps[0][0] * (1.0 - xratio) * yratio
                            xsin = xsin + complex_comps[0][1] * (1.0 - xratio) * yratio
                            denom = denom + (1.0 - xratio) * yratio
                        if not numpy.isnan(xhi_yhi_amp) and not numpy.isnan(xhi_yhi_phase):
                            xcos = xcos + complex_comps[1][0] * xratio * yratio
                            xsin = xsin + complex_comps[1][1] * xratio * yratio
                            denom = denom + xratio * yratio
                        if not numpy.isnan(xlo_ylo_amp) and not numpy.isnan(xlo_ylo_phase):
                            xcos = xcos + complex_comps[2][0] * (1.0 - xratio) * (1 - yratio)
                            xsin = xsin + complex_comps[2][1] * (1.0 - xratio) * (1 - yratio)
                            denom = denom + (1.0 - xratio) * (1.0 - yratio)
                        if not numpy.isnan(xhi_ylo_amp) and not numpy.isnan(xhi_ylo_phase):
                            xcos = xcos + complex_comps[3][0] * (1.0 - yratio) * xratio
                            xsin = xsin + complex_comps[3][1] * (1.0 - yratio) * xratio
                            denom = denom + (1.0 - yratio) * xratio
                        xcos = xcos / denom
                        xsin = xsin / denom
                        amp = math.sqrt(xcos * xcos + xsin * xsin)

                        # Compute interpolated phase
                        phase = math.degrees(math.acos(xcos / amp))
                        amp /= 100.0
                        if xsin < 0.0:
                            phase = 360.0 - phase
                        phase += (360. if positive_ph and phase < 0 else 0)
                        speed = NOAA_SPEEDS[con][0] if con in NOAA_SPEEDS else numpy.nan
                        rows[i][con] = (amp, phase, speed)

        self.data = [
            pd.DataFrame.from_dict(row, orient='index', columns=['amplitude', 'phase', 'speed'])
            for row in rows
        ]
        return self



"""Class to represent tides in harmonica."""

# 1. Standard Python modules
from datetime import datetime

# 2. Third party modules
import numpy as np
import pandas as pd
from pytides.astro import astro
import pytides.constituent as pycons
from pytides.tide import Tide as pyTide

# 3. Aquaveo modules

# 4. Local modules
from .resource import ResourceManager
from .tidal_constituents import Constituents
from .tidal_database import NOAA_SPEEDS


class Tide:
    """Harmonica tide object."""

    # Dictionary to convert generic uppercase constituent name to pytides name;
    # if name isn't listed, then the associated pytides name is all uppercase
    PYTIDES_CON_MAPPER = {
        'SA': 'Sa',
        'SSA': 'Ssa',
        'MM': 'Mm',
        'MF': 'Mf',
        'NU2': 'nu2',
        'LAMBDA2': 'lambda2',
        'RHO1': 'rho1',
        'MU2': 'mu2',
    }

    def __init__(self, model=ResourceManager.DEFAULT_RESOURCE):
        """Constructor.

        Args:
            model (str): Name of the model to use. See ResourceManager constants for valid values.
        """
        # tide dataframe:
        #   date_times (year, month, day, hour, minute, second; UTC/GMT)
        self.data = pd.DataFrame(columns=['datetimes', 'water_level'])
        self.constituents = Constituents(model=model)

    def reconstruct_tide(self, loc, times, model=None, cons=None, positive_ph=False, offset=None):
        """Rescontruct a tide signal water levels at the given location and times.

        Args:
            loc (tuple(float, float)): latitude [-90, 90] and longitude [-180 180] or [0 360] of the requested point.
            times (ndarray(datetime)): Array of datetime objects associated with each water level data point.
            model (str, optional): Model name, defaults to 'tpxo8'.
            cons (list(str), optional): List of constituents requested, defaults to all constituents if None or empty.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).
            offset (float, optional): If not None, includes a generic constituent with a phase of the given value.
        """
        # get constituent information
        cons = cons if cons else []
        self.constituents.get_components([loc], cons, positive_ph, model=model)

        ncons = len(self.constituents.data[0]) + (1 if offset is not None else 0)
        tide_model = np.zeros(ncons, dtype=pyTide.dtype)
        # load specified model constituent components into pytides model object
        for i, key in enumerate(self.constituents.data[0].index.values):
            tide_model[i]['constituent'] = eval('pycons._{}'.format(self.PYTIDES_CON_MAPPER.get(key, key)))
            tide_model[i]['amplitude'] = self.constituents.data[0].loc[key].amplitude
            tide_model[i]['phase'] = self.constituents.data[0].loc[key].phase
        # if an offset is provided then add as spoofed constituent Z0
        if offset is not None:
            tide_model[-1]['constituent'] = pycons._Z0
            tide_model[-1]['amplitude'] = 0.
            tide_model[-1]['phase'] = offset

        # reconstruct the tides, store in self
        self.data['datetimes'] = pd.Series(times)
        self.data['water_level'] = pd.Series(pyTide(model=tide_model, radians=False).at(times), index=self.data.index)

        return self

    def deconstruct_tide(self, water_level, times, cons=None, n_period=6, positive_ph=False):
        """Method to use pytides to deconstruct the tides and reorganize results back into the class structure.

        Args:
            water_level (ndarray(float)): Array of water levels.
            times (ndarray(datetime)): Array of datetime objects associated with each water level data point.
            cons (list(str), optional): List of constituents requested, defaults to all constituents if None or empty.
            n_period(int): Number of periods a constituent must complete during times to be considered in analysis.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).

        Returns:
            A dataframe of constituents information in Constituents class
        """
        # Fit the tidal data to the harmonic model using pytides
        if not cons:
            cons = pycons.noaa
        else:
            cons = [eval('pycons._{}'.format(self.PYTIDES_CON_MAPPER.get(c, c))) for c in cons if c in NOAA_SPEEDS]
        self.model_to_dataframe(pyTide.decompose(water_level, times, constituents=cons, n_period=n_period), times[0],
                                positive_ph=positive_ph)
        return self

    def model_to_dataframe(self, tide, t0=None, positive_ph=False):
        """Method to reorganize data from the pytides tide model format into the native dataframe format.

        Args:
            tide (pytides object): Tide model object from pytides.
            t0 (datetime, optional): Time at which to evaluate speed based on astronomical parameters (speeds vary
                slowly over time), defaults to current date and time.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).

        Returns:
            A dataframe of constituents information in Constituents class
        """
        t0 = t0 if t0 is not None else datetime.now()

        # helper function to extract constituent information from Tide model
        def extractor(c):
            # info: name, amplitude, phase, speed
            return c[0].name.upper(), c[1], c[2], c[0].speed(astro(t0))

        # create a filtered array of constituent information
        cons = np.asarray(np.vectorize(extractor)(tide.model[tide.model['constituent'] != pycons._Z0])).T
        # convert into dataframe
        df = pd.DataFrame(cons[:, 1:], index=cons[:, 0], columns=['amplitude', 'phase', 'speed'], dtype=float)
        self.constituents.data[0] = pd.concat([self.constituents.data[0], df], axis=0, join='inner')
        # convert phase if necessary
        if not positive_ph:
            self.constituents.data[0]['phase'] = np.where(self.constituents.data[0]['phase'] > 180.,
                                                          self.constituents.data[0]['phase'] - 360.,
                                                          self.constituents.data[0]['phase'])



"""Class to manage the TPXO tidal database models."""

# 1. Standard Python modules

# 2. Third party modules
import numpy as np
import pandas as pd

# 3. Aquaveo modules

# 4. Local modules
from .resource import ResourceManager
from .tidal_database import NOAA_SPEEDS, TidalDB


DEFAULT_TPXO_RESOURCE = 'tpxo10_atlas'  # was 'tpxo9'


class TpxoDB(TidalDB):
    """Harmonica tidal constituents."""

    def __init__(self, model=DEFAULT_TPXO_RESOURCE):
        """Constructor for the TPXO tidal extractor.

        Args:
            model (:obj:`str`, optional): The name of the TPXO model. See resource.py for supported models.

        """
        model = model.lower()  # Be case-insensitive
        if model not in ResourceManager.TPXO_MODELS:  # Check for valid TPXO model
            raise ValueError("\'{}\' is not a supported TPXO model. Must be one of: {}.".format(
                model, ", ".join(ResourceManager.TPXO_MODELS).strip()
            ))
        super().__init__(model)

    def get_components(self, locs, cons=None, positive_ph=False):
        """Get the amplitude, phase, and speed of specified constituents at specified point locations.

        Args:
            locs (:obj:`list` of :obj:`tuple` of :obj:`float`): latitude [-90, 90] and longitude [-180 180] or [0 360]
                of the requested points.
            cons (:obj:`list` of :obj:`str`, optional): List of the constituent names to get amplitude and phase for. If
                not supplied, all valid constituents will be extracted.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).

        Returns:
           :obj:`list` of :obj:`pandas.DataFrame`: A list of dataframes of constituent information including
                amplitude (meters), phase (degrees) and speed (degrees/hour, UTC/GMT). The list is parallel with locs,
                where each element in the return list is the constituent data for the corresponding element in locs.
                Empty list on error. Note that function uses fluent interface pattern.

        """
        n_locs = len(locs)
        # Accumulate each point's constituent rows, then build one DataFrame per point at the end. This avoids the
        # per-cell DataFrame.loc writes the old inner loop paid for every constituent of every point.
        rows = [{} for _ in range(n_locs)]

        # if no constituents were requested, return all available
        if cons is None or not len(cons):
            cons = list(self.resources.available_constituents())
        requested = set(cons)
        units_multiplier = self.resources.get_units_multiplier()

        # Requested coordinates as arrays; longitudes normalized to [0, 360) exactly as the scalar path did.
        lats = np.array([loc[0] for loc in locs], dtype=float)
        lons = np.array([loc[1] for loc in locs], dtype=float)
        lons = np.where(lons < 0.0, lons + 360.0, lons)

        # open the netcdf database(s)
        single_file = self.resources.model_atts.is_consolidated_file
        for d in self.resources.get_datasets(cons):
            for dset in d:
                # remove unnecessary data array dimensions if present (e.g. tpxo9)
                if 'nx' in dset.lat_z.dims:
                    dset['lat_z'] = dset.lat_z.sel(nx=0, drop=True)
                if 'ny' in dset.lon_z.dims:
                    dset['lon_z'] = dset.lon_z.sel(ny=0, drop=True)
                # get the dataset constituent name array from data cube
                if single_file:
                    nc_names = [x.tobytes().decode('utf-8').strip().upper() for x in dset.con.values]
                else:
                    nc_names = [dset.con.item().decode('utf-8').strip().upper()]

                # Bounding indices for every point at once. bisect(a, x) == np.searchsorted(a, x, side='right'),
                # so the interior indices match the old per-point scalar path.
                lon_z = dset.lon_z.values
                lat_z = dset.lat_z.values
                right = np.searchsorted(lon_z, lons, side='right')
                top = np.searchsorted(lat_z, lats, side='right')
                # A point is in the grid only if its full 2x2 window is in range. Out-of-domain points get NaN
                # rows (as the LeProvost/ADCIRC extractors do) instead of raising and aborting the whole batch.
                in_domain = (right >= 1) & (right < len(lon_z)) & (top >= 1) & (top < len(lat_z))
                # Clip so the vectorized weight math never indexes out of bounds for an out-of-domain point; its
                # weights are discarded (its data window stays NaN). In-domain indices are unchanged by the clip.
                right = np.clip(right, 1, len(lon_z) - 1)
                left = right - 1
                top = np.clip(top, 1, len(lat_z) - 1)
                bottom = top - 1
                # Bilinear spline weights per point, shaped (n_locs, 2, 2) to line up with each 2x2 data window:
                # row = lon (left, right), col = lat (bottom, top). Same layout the old scalar code used.
                dx = (lons - lon_z[left]) / (lon_z[right] - lon_z[left])
                dy = (lats - lat_z[bottom]) / (lat_z[top] - lat_z[bottom])
                weights = np.stack([
                    (1. - dx) * (1. - dy),  # left, bottom
                    (1. - dx) * dy,         # left, top
                    dx * (1. - dy),         # right, bottom
                    dx * dy,                # right, top
                ], axis=-1).reshape(n_locs, 2, 2)
                weights = weights / weights.sum(axis=(1, 2), keepdims=True)

                for c in requested & set(nc_names):
                    con_idx = nc_names.index(c) if single_file else 0
                    # Read only each in-domain point's 2x2 window from the lazily-opened arrays (never the whole
                    # grid); out-of-domain points keep NaN so they interpolate to NaN.
                    re_block = np.full((n_locs, 2, 2), np.nan)
                    im_block = np.full((n_locs, 2, 2), np.nan)
                    for i in np.nonzero(in_domain)[0]:
                        if single_file:
                            query = np.s_[con_idx, left[i]:right[i] + 1, bottom[i]:top[i] + 1]
                        else:
                            query = np.s_[left[i]:right[i] + 1, bottom[i]:top[i] + 1]
                        re_block[i] = dset.hRe[query].values
                        im_block[i] = dset.hIm[query].values
                    # weighted tide from the real and imaginary components, vectorized over points
                    real = (re_block * weights).sum(axis=(1, 2))
                    imag = -(im_block * weights).sum(axis=(1, 2))
                    h = real + 1j * imag
                    phase = np.angle(h, deg=True)
                    if positive_ph:
                        phase = np.where(phase < 0.0, phase + 360.0, phase)
                    amplitude = np.absolute(h) * units_multiplier
                    speed = np.where(in_domain, NOAA_SPEEDS[c][0], np.nan)
                    for i in range(n_locs):
                        rows[i][c] = (float(amplitude[i]), float(phase[i]), float(speed[i]))

        self.data = [
            pd.DataFrame.from_dict(row, orient='index', columns=['amplitude', 'phase', 'speed'])
            for row in rows
        ]
        return self



"""The deconstruct CLI command."""

# 1. Standard Python modules
import argparse
import sys

# 2. Third party modules
import pandas as pd

# 3. Aquaveo modules

# 4. Local modules
from .common import add_common_args, add_const_out_args
from .. import harmonica


DESCR = 'Deconstruct the signal into its tidal constituents.'
EXAMPLE = """
Example:

    harmonica deconstruct CO-OPS__8760922__wl.csv --columns "Date Time" "Water Level" \
        --datetime_format '%Y-%m-%d %H:%M' -C M2 S2 N2 K1
"""


def config_parser(p, sub=False):
    """Configure the command line arguments passed the deconstruct CLI command.

    Args:
        p (ArgumentParser): The argument parser
        sub (Optional[bool]): True if this is a resources subparser
    """
    # Subparser info
    if sub:
        p = p.add_parser(
            'deconstruct',
            description=DESCR,
            help=DESCR,
            epilog=EXAMPLE,
            add_help=False,
        )

    # Required positional arguments
    p.add_argument(
        'signal',
        type=str,
        help='File to read the tidal signal to deconstruct (CSV formatted)',
        metavar='SIGNAL',
    )

    add_common_args(p)
    p.add_argument(
        '--datetime_format',
        default='%Y-%m-%d %H:%M:%S',
        help="Format of 'datetime' values in signal file, default: '%%Y-%%m-%%d %%H:%%M:%%S' (used by Pandas "
             "datetime parser)",
        dest='dt_format',
    )
    p.add_argument(
        '--columns',
        nargs='+',
        default=[0, 1],
        help="Name or index of columns in signal file to extract times and water levels; multiple columns can be "
             "specified to combine and parse as a datetime; the last column specified is assumed to be water levels, "
             "default: '0 1'",
        dest='dt_cols',
        metavar='COL',
    )
    p.add_argument(
        '--header',
        default=0,
        help="Row number that contains the column names (specify 'None' if no column names), default: '0'",
    )
    p.add_argument(
        '--sep',
        default=',',
        help="Field delimiter of the signal file, default: ',' (comma)",
    )
    add_const_out_args(p)
    p.add_argument(
        '--num_periods',
        type=int,
        default=6,
        help="Number of periods a constituent must complete during signal length to be considered, default: 6",
    )


def parse_args(args):
    """Parse the command line arguments passed the deconstruct CLI command.

    Args:
        args (...): Variable length positional arguments

    Returns:
        ArgumentParser: The command line argument parser
    """
    p = argparse.ArgumentParser(
        description=DESCR,
        epilog=EXAMPLE,
        add_help=False,
    )
    config_parser(p)

    return p.parse_args(args)


def execute(args):
    """Execute the deconstruct CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    args.dt_cols = [int(x) if isinstance(x, str) and x.isdigit() else x for x in args.dt_cols]
    opts = {
        'date_parser': lambda x: [pd.datetime.strptime(d, args.dt_format) for d in x],
        'parse_dates': {'datetimes': args.dt_cols[:-1]},
        'header': args.header,
        'sep': args.sep,
    }
    try:
        df = pd.read_csv(args.signal, **opts)
    except ValueError as e:
        if 'not in list' in str(e):
            print("\nThe column name '{}' is not recognized.".format(str(e).split("'")[1]))
        elif 'match format' in str(e):
            print("\nThe signal's datetime does not match the given format: {}. Verify format and/or header row "
                  "number".format(args.dt_format))
        else:
            print(str(e))
    except KeyError:
        print("\nThe column name(s) to parse the signal's datetime are not recognized.")

    wl = args.dt_cols[-1]
    wl = df.columns[wl] if isinstance(wl, int) else wl
    try:
        tide = harmonica.Tide().deconstruct_tide(df[wl], df['datetimes'], cons=args.cons,
                                                 n_period=args.num_periods, positive_ph=args.positive_phase)
    except RuntimeWarning as w:
        if 'Number of calls to function has reached maxfev' in str(w):
            print("\nThe solver failed to converge to a solution. Provide a longer signal.")
        else:
            print(str(w))
        print('\nFailed.\n')
        return

    out = tide.constituents.data.to_csv(args.output, sep='\t', header=True, index=True, index_label='constituent')
    if args.output is None:
        print(out)
    print('\nComplete.\n')


def main(args=None):
    """Entry point for the deconstruct CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    if not args:
        args = sys.argv[1:]
    try:
        execute(parse_args(args))
    except RuntimeError as e:
        print(str(e))
        sys.exit(1)
    return



"""The download CLI command."""

# 1. Standard Python modules
import argparse
import sys

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from ..resource import ResourceManager


DESCR = 'Download and pre-position model resources for subsequent analysis calls.'
EXAMPLE = """
Example:

    harmonica download tpxo8
"""


def config_parser(p, sub=False):
    """Configure the command line arguments passed the download CLI command.

    Args:
        p (ArgumentParser): The argument parser
        sub (Optional[bool]): True if this is a resources subparser
    """
    # Subparser info
    if sub:
        p = p.add_parser(
            'download',
            description=DESCR,
            help=DESCR,
            epilog=EXAMPLE,
            add_help=False,
        )

    p.add_argument(
        'model',
        choices=ResourceManager.RESOURCES.keys(),
        default=ResourceManager.DEFAULT_RESOURCE,
        help='Constituent model specification, default: tpxo8',
    )


def parse_args(args):
    """Parse the command line arguments passed the download CLI command.

    Args:
        args (...): Variable length positional arguments

    Returns:
        ArgumentParser: The command line argument parser
    """
    p = argparse.ArgumentParser(
        description=DESCR,
        epilog=EXAMPLE,
        add_help=False,
    )
    config_parser(p)
    return p.parse_args(args)


def execute(args):
    """Execute the download CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    ResourceManager(model=args.model).download_model()
    print('\nComplete.\n')


def main(args=None):
    """Entry point for the download CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    if not args:
        args = sys.argv[1:]
    try:
        execute(parse_args(args))
    except RuntimeError as e:
        print(str(e))
        sys.exit(1)
    return



"""Resource managers for the tidal database models supported in harmonica."""

# 1. Standard Python modules
from abc import ABCMeta, abstractmethod
import os
import shutil
import urllib.request
from zipfile import ZipFile

# 2. Third party modules
import xarray as xr

# 3. Aquaveo modules

# 4. Local modules
from harmonica import config


MAX_NUM_CONS = 37  # Maximum number of constituents in all available models


class Resources(metaclass=ABCMeta):
    """Abstract base class for model resources."""

    # Default: data is one file per constituent. Override to True for single-file models.
    # Consumed by TpxoDB.get_components to dispatch between the per-con and consolidated code paths.
    is_consolidated_file = False
    # Default: the on-disk data subdirectory name (used under both the %APPDATA% data dir and
    # pre_existing_data_dir) matches the model name. Override (e.g. 'tpxo9_atlas_v5') when it is version-suffixed.
    data_dir_name = None

    @abstractmethod
    def resource_attributes(self):
        """Get the resource attributes of a model (e.g. web url, compression type).

        Returns:
            dict: Dictionary of model resource attributes
        """
        return {}

    @abstractmethod
    def dataset_attributes(self):
        """Get the dataset attributes of a model (e.g. unit multiplier, grid dimensions).

        Returns:
            dict: Dictionary of model dataset attributes
        """
        return {}

    @abstractmethod
    def available_constituents(self):
        """Get all the available constituents of a model.

        Returns:
            list: List of all the available constituents
        """
        return []

    @abstractmethod
    def constituent_groups(self):
        """Get all the available constituents of a model grouped by compatible file types.

        Returns:
            list[list]: 2-D list of available constituents, where the first dimension groups compatible files
        """
        return []

    @abstractmethod
    def constituent_resource(self, con):
        """Get the resource name of a constituent.

        Returns:
            str: Name of the constituent's resource
        """
        return None


class Tpxo8Resources(Resources):
    """TPXO8 resources."""
    TPXO8_CONS = [
        {  # 1/30 degree
            'K1': 'hf.k1_tpxo8_atlas_30c_v1.nc',
            'K2': 'hf.k2_tpxo8_atlas_30c_v1.nc',
            'M2': 'hf.m2_tpxo8_atlas_30c_v1.nc',
            'M4': 'hf.m4_tpxo8_atlas_30c_v1.nc',
            'N2': 'hf.n2_tpxo8_atlas_30c_v1.nc',
            'O1': 'hf.o1_tpxo8_atlas_30c_v1.nc',
            'P1': 'hf.p1_tpxo8_atlas_30c_v1.nc',
            'Q1': 'hf.q1_tpxo8_atlas_30c_v1.nc',
            'S2': 'hf.s2_tpxo8_atlas_30c_v1.nc',
        },
        {  # 1/6 degree
            'MF': 'hf.mf_tpxo8_atlas_6.nc',
            'MM': 'hf.mm_tpxo8_atlas_6.nc',
            'MN4': 'hf.mn4_tpxo8_atlas_6.nc',
            'MS4': 'hf.ms4_tpxo8_atlas_6.nc',
        },
    ]

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Returns a dict of the resource attributes that are disabled (TPXO8 is not freely available)."""
        return {
            'url': None,  # Resources must already exist. Licensing restrictions prevent hosting files.
            'archive': None,
        }

    def dataset_attributes(self):
        """Returns a dict of the TPXO8 dataset attributes (currently just 'units_muliplier')."""
        return {
            'units_multiplier': 0.001,  # mm to meter
        }

    def available_constituents(self):
        """Returns a list of the constituents supported by the TPXO8 tidal database."""
        # get keys from const groups as list of lists and flatten
        return [c for sl in [grp.keys() for grp in self.TPXO8_CONS] for c in sl]

    def constituent_groups(self):
        """Returns a list of the constituents supported by the TPXO8 tidal database (2-D)."""
        return [self.TPXO8_CONS[0], self.TPXO8_CONS[1]]

    def constituent_resource(self, con):
        """Get the filename given a constituent.

        Args:
            con (str): Name of the constituent to get resource filename of. See TPXO8_CONS for list of available
                constituents

        Returns:
            str: The basename of the resource file if the specified constituent is supported by the TPXO8 tidal
                database, else None
        """
        con = con.upper()
        for group in self.TPXO8_CONS:
            if con in group:
                return group[con]
        return None


class Tpxo9Resources(Resources):
    """TPXO9 resources."""
    TPXO9_CONS = {'2N2', 'K1', 'K2', 'M2', 'M4', 'MF', 'MM', 'MN4', 'MS4', 'N2', 'O1', 'P1', 'Q1', 'S1', 'S2'}
    DEFAULT_RESOURCE_FILE = 'tpxo9_netcdf/h_tpxo9.v1.nc'
    is_consolidated_file = True

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Returns a dict of the resource attributes that are disabled (TPXO9 is not freely available)."""
        return {
            'url': None,  # Resources must already exist. Licensing restrictions prevent hosting files.
            'archive': 'gz',
        }

    def dataset_attributes(self):
        """Returns a dict of the TPXO9 dataset attributes (currently just 'units_muliplier')."""
        return {
            'units_multiplier': 1.0,  # meter
        }

    def available_constituents(self):
        """Returns a list of the constituents supported by the TPXO9 tidal database."""
        return self.TPXO9_CONS

    def constituent_groups(self):
        """Returns a list of the constituents supported by the TPXO9 tidal database (2-D)."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """Get the filename given a constituent.

        Args:
            con (str): Name of the constituent to get resource filename of. See TPXO9_CONS for list of available
                constituents

        Returns:
            str: The basename of the resource file if the specified constituent is supported by the TPXO9 tidal
                database, else None
        """
        if con.upper() in self.TPXO9_CONS:
            return self.DEFAULT_RESOURCE_FILE
        else:
            return None


class Tpxo9AtlasResources(Resources):
    """TPXO9-atlas-v5 resources (1/30 degree global atlas, per-constituent files)."""
    TPXO9_ATLAS_CONS = {
        '2N2': 'h_2n2_tpxo9_atlas_30_v5.nc',
        'K1': 'h_k1_tpxo9_atlas_30_v5.nc',
        'K2': 'h_k2_tpxo9_atlas_30_v5.nc',
        'M2': 'h_m2_tpxo9_atlas_30_v5.nc',
        'M4': 'h_m4_tpxo9_atlas_30_v5.nc',
        'MF': 'h_mf_tpxo9_atlas_30_v5.nc',
        'MM': 'h_mm_tpxo9_atlas_30_v5.nc',
        'MN4': 'h_mn4_tpxo9_atlas_30_v5.nc',
        'MS4': 'h_ms4_tpxo9_atlas_30_v5.nc',
        'N2': 'h_n2_tpxo9_atlas_30_v5.nc',
        'O1': 'h_o1_tpxo9_atlas_30_v5.nc',
        'P1': 'h_p1_tpxo9_atlas_30_v5.nc',
        'Q1': 'h_q1_tpxo9_atlas_30_v5.nc',
        'S1': 'h_s1_tpxo9_atlas_30_v5.nc',
        'S2': 'h_s2_tpxo9_atlas_30_v5.nc',
    }
    is_consolidated_file = False
    data_dir_name = 'tpxo9_atlas_v5'

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Disabled (TPXO9-atlas-v5 is licensed; registration required, no free distribution)."""
        return {
            'url': None,  # Resources must already exist. Licensing restrictions prevent hosting files.
            'archive': None,  # OSU ships TPXO9-atlas-v5 as a plain directory; no archive wrapper.
        }

    def dataset_attributes(self):
        """Dataset attributes (mm storage, scale to m)."""
        return {
            'units_multiplier': 0.001,
        }

    def available_constituents(self):
        """The 15 constituents in TPXO9-atlas-v5."""
        return list(self.TPXO9_ATLAS_CONS.keys())

    def constituent_groups(self):
        """Single uniform-resolution group."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """Map constituent name to per-con filename, or None if unsupported."""
        return self.TPXO9_ATLAS_CONS.get(con.upper())


class Tpxo10Resources(Resources):
    """TPXO10v2 resources (1/6 degree global, consolidated single-file layout)."""
    TPXO10_CONS = {
        'M2', 'S2', 'N2', 'K2', 'K1', 'O1', 'P1', 'Q1', 'MM', 'MF',
        'MSF', 'M4', 'MN4', 'MS4', '2N2', 'S1', '2Q1', 'J1', 'L2', 'M3',
        'MU2', 'NU2', 'OO1', 'T2', 'M1',
    }
    DEFAULT_RESOURCE_FILE = 'h_tpxo10.v2.nc'
    is_consolidated_file = True
    data_dir_name = 'tpxo10v2'

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Disabled (TPXO10v2 is licensed; registration required, no free distribution)."""
        return {
            'url': None,  # Resources must already exist. Licensing restrictions prevent hosting files.
            'archive': None,  # OSU ships TPXO10v2 as a plain directory; no archive wrapper.
        }

    def dataset_attributes(self):
        """Dataset attributes (m storage, no scaling)."""
        return {
            'units_multiplier': 1.0,
        }

    def available_constituents(self):
        """The 25 constituents in TPXO10v2."""
        return self.TPXO10_CONS

    def constituent_groups(self):
        """Single uniform-resolution group."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """All supported cons live in the consolidated file."""
        if con.upper() in self.TPXO10_CONS:
            return self.DEFAULT_RESOURCE_FILE
        return None


class Tpxo10AtlasResources(Resources):
    """TPXO10-atlas-v2 resources (1/30 degree global atlas, per-constituent files)."""
    TPXO10_ATLAS_CONS = {
        '2N2': 'h_2n2_tpxo10_atlas_30_v2.nc',
        'K1': 'h_k1_tpxo10_atlas_30_v2.nc',
        'K2': 'h_k2_tpxo10_atlas_30_v2.nc',
        'M2': 'h_m2_tpxo10_atlas_30_v2.nc',
        'M4': 'h_m4_tpxo10_atlas_30_v2.nc',
        'MF': 'h_mf_tpxo10_atlas_30_v2.nc',
        'MM': 'h_mm_tpxo10_atlas_30_v2.nc',
        'MN4': 'h_mn4_tpxo10_atlas_30_v2.nc',
        'MS4': 'h_ms4_tpxo10_atlas_30_v2.nc',
        'N2': 'h_n2_tpxo10_atlas_30_v2.nc',
        'O1': 'h_o1_tpxo10_atlas_30_v2.nc',
        'P1': 'h_p1_tpxo10_atlas_30_v2.nc',
        'Q1': 'h_q1_tpxo10_atlas_30_v2.nc',
        'S1': 'h_s1_tpxo10_atlas_30_v2.nc',
        'S2': 'h_s2_tpxo10_atlas_30_v2.nc',
    }
    is_consolidated_file = False
    data_dir_name = 'tpxo10_atlas_v2'

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Disabled (licensed; registration required)."""
        return {
            'url': None,  # Resources must already exist. Licensing restrictions prevent hosting files.
            'archive': None,  # OSU ships TPXO10-atlas-v2 as a plain directory; no archive wrapper.
        }

    def dataset_attributes(self):
        """Dataset attributes (mm storage)."""
        return {
            'units_multiplier': 0.001,
        }

    def available_constituents(self):
        """The 15 constituents in TPXO10-atlas-v2."""
        return list(self.TPXO10_ATLAS_CONS.keys())

    def constituent_groups(self):
        """Single uniform-resolution group."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """Map constituent name to per-con filename, or None if unsupported."""
        return self.TPXO10_ATLAS_CONS.get(con.upper())


class LeProvostResources(Resources):
    """LeProvost resources."""
    LEPROVOST_CONS = {'K1', 'K2', 'M2', 'N2', 'O1', 'P1', 'Q1', 'S2', 'NU2', 'MU2', '2N2', 'T2', 'L2'}
    DEFAULT_RESOURCE_FILE = 'leprovost_tidal_db.nc'

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Returns a dict of the resource attributes needed to fetch the freely available LeProvost tidal database."""
        return {
            'url': 'http://sms.aquaveo.com/leprovost_tidal_db.zip',
            'archive': 'zip',  # zip compression
        }

    def dataset_attributes(self):
        """Returns a dict of the LeProvost dataset attributes."""
        return {
            'units_multiplier': 1.0,  # meter
            'num_lats': 361,
            'num_lons': 720,
            'min_lon': -180.0,
        }

    def available_constituents(self):
        """Returns a list of the constituents supported by the LeProvost tidal database."""
        return self.LEPROVOST_CONS

    def constituent_groups(self):
        """Returns a list of the constituents supported by the LeProvost tidal database (2-D)."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """Get the filename given a constituent.

        Args:
            con (str): Name of the constituent to get resource filename of. See LEPROVOST_CONS for list of available
                constituents

        Returns:
            str: The basename of the resource file if the specified constituent is supported by the LeProvost tidal
                database, else None
        """
        if con.upper() in self.LEPROVOST_CONS:
            return self.DEFAULT_RESOURCE_FILE
        else:
            return None


class FES2014Resources(Resources):
    """FES2014 resources."""
    FES2014_CONS = {
        '2N2': '2n2.nc',
        'EPS2': 'eps2.nc',
        'J1': 'j1.nc',
        'K1': 'k1.nc',
        'K2': 'k2.nc',
        'L2': 'l2.nc',
        'LA2': 'la2.nc',
        'M2': 'm2.nc',
        'M3': 'm3.nc',
        'M4': 'm4.nc',
        'M6': 'm6.nc',
        'M8': 'm8.nc',
        'MF': 'mf.nc',
        'MKS2': 'mks2.nc',
        'MM': 'mm.nc',
        'MN4': 'mn4.nc',
        'MS4': 'ms4.nc',
        'MSF': 'msf.nc',
        'MSQM': 'msqm.nc',
        'MTM': 'mtm.nc',
        'MU2': 'mu2.nc',
        'N2': 'n2.nc',
        'N4': 'n4.nc',
        'NU2': 'nu2.nc',
        'O1': 'o1.nc',
        'P1': 'p1.nc',
        'Q1': 'q1.nc',
        'R2': 'r2.nc',
        'S1': 's1.nc',
        'S2': 's2.nc',
        'S4': 's4.nc',
        'SA': 'sa.nc',
        'SSA': 'ssa.nc',
        'T2': 't2.nc',
    }

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Returns a dict of the resource attributes that are disabled (FES2014 is not freely available)."""
        return {
            'url': None,  # Resources must already exist. Licensing restrictions prevent hosting files.
            'archive': None,
        }

    def dataset_attributes(self):
        """Returns a dict of the FES2014 dataset attributes."""
        return {
            'units_multiplier': 1.0,  # meter
            'num_lats': 2881,
            'num_lons': 5760,
            'min_lon': 0.0,
        }

    def available_constituents(self):
        """Returns a list of the constituents supported by the FES2014 tidal database."""
        return self.FES2014_CONS.keys()

    def constituent_groups(self):
        """Returns a list of the constituents supported by the FES2014 tidal database (2-D)."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """Get the filename given a constituent.

        Args:
            con (str): Name of the constituent to get resource filename of. See FES2014_CONS for list of available
                constituents

        Returns:
            str: The basename of the resource file if the specified constituent is supported by the FES2014 tidal
                database, else None
        """
        con = con.upper()
        if con in self.FES2014_CONS:
            return self.FES2014_CONS[con]
        else:
            return None


class Adcirc2015Resources(Resources):
    """ADCIRC (v2015) resources."""
    ADCIRC_CONS = {
        'M2', 'S2', 'N2', 'K1', 'M4', 'O1', 'M6', 'Q1', 'K2', 'L2', '2N2', 'R2', 'T2', 'LAMBDA2', 'MU2',
        'NU2', 'J1', 'M1', 'OO1', 'P1', '2Q1', 'RHO1', 'M8', 'S4', 'S6', 'M3', 'S1', 'MK3', '2MK3', 'MN4',
        'MS4', '2SM2', 'MF', 'MSF', 'MM', 'SA', 'SSA'
    }
    DEFAULT_RESOURCE_FILE = 'all_adcirc.nc'

    def __init__(self):
        """Constructor."""
        super().__init__()

    def resource_attributes(self):
        """Returns a dict of the resource attributes needed to fetch the freely available ADCIRC tidal database."""
        return {
            'url': 'http://sms.aquaveo.com/',
            'archive': None,  # Uncompressed NetCDF file
        }

    def dataset_attributes(self):
        """Returns a dict of the dataset attributes (currently only 'units_multiplier')."""
        return {'units_multiplier': 1.0}  # meter

    def available_constituents(self):
        """Returns a list of the constituents supported by the ADCIRC tidal database."""
        return self.ADCIRC_CONS

    def constituent_groups(self):
        """Returns a list of the constituents supported by the ADCIRC tidal database (2-D)."""
        return [self.available_constituents()]

    def constituent_resource(self, con):
        """Get the filename given a constituent.

        Args:
            con (str): Name of the constituent to get resource filename of. See ADCIRC_CONS for list of available
                constituents

        Returns:
            str: The basename of the resource file if the specified constituent is supported by the ADCIRC tidal
                database, else None
        """
        if con.upper() in self.ADCIRC_CONS:
            return self.DEFAULT_RESOURCE_FILE
        else:
            return None


class ResourceManager(object):
    """Harmonica resource manager to retrieve and access tide models."""

    RESOURCES = {
        'tpxo8': Tpxo8Resources(),
        'tpxo9': Tpxo9Resources(),
        'tpxo9_atlas': Tpxo9AtlasResources(),
        'tpxo10': Tpxo10Resources(),
        'tpxo10_atlas': Tpxo10AtlasResources(),
        'leprovost': LeProvostResources(),
        'fes2014': FES2014Resources(),
        'adcirc2015': Adcirc2015Resources(),
    }
    TPXO_MODELS = {'tpxo8', 'tpxo9', 'tpxo9_atlas', 'tpxo10', 'tpxo10_atlas'}
    LEPROVOST_MODELS = {'fes2014', 'leprovost'}
    ADCIRC_MODELS = {'adcirc2015'}
    DEFAULT_RESOURCE = 'tpxo10_atlas'  # was 'tpxo9'

    def __init__(self, model=DEFAULT_RESOURCE):
        """Constructor.

        Args:
            model (str): Name of the model to use initially.  See the constants defined in ResourceManager for valid
                values.
        """
        if model not in self.RESOURCES:
            raise ValueError('Model not recognized.')
        self.model = model
        self.model_atts = self.RESOURCES[self.model]
        self.datasets = []

    def __del__(self):
        """Deleter - closes Dataset file handles."""
        for d in self.datasets:
            for dset in d:
                dset.close()

    @staticmethod
    def data_dir_exists(model):
        """Check if a model's data directory exists in either the default location or the configurable one.

        Args:
            model (str): Name of the model. See the constants defined in ResourceManager for valid values.

        Returns:
            bool: True if the model's data folder exists in either location.
        """
        resource = ResourceManager.RESOURCES.get(model)
        data_dir = (resource.data_dir_name if resource is not None else None) or model
        if os.path.isdir(os.path.join(config['data_dir'], data_dir)):
            return True  # Exists in the default %APPDATA% folder
        if os.path.isdir(os.path.join(config['pre_existing_data_dir'], data_dir)):
            return True  # Exists in the user configurable folder
        return False

    @staticmethod
    def available_models():
        """Get a dict of flags indicating whether a tidal model has been installed.

        This just performs a quick check for the existence of model data folders. If you need a more robust
        check, write something else. Needed something fast for dialogs.

        Returns:
            dict: {'model': True if available else False}
        """
        return {model: ResourceManager.data_dir_exists(model) for model in ResourceManager.RESOURCES}

    def available_constituents(self):
        """Returns a list of the available constituents for the current model."""
        return self.model_atts.available_constituents()

    def get_units_multiplier(self):
        """Returns the units multiplier for the current model."""
        return self.model_atts.dataset_attributes()['units_multiplier']

    def download(self, resource, destination_dir):
        """Download a specified model resource."""
        if not os.path.isdir(destination_dir):
            os.makedirs(destination_dir)

        rsrc_atts = self.model_atts.resource_attributes()
        url = rsrc_atts['url']
        # Check if we can download resources for this model.
        if url is None:
            raise ValueError("Automatic fetching of resources is not available for the {} model.".format(self.model))

        if rsrc_atts['archive'] is None:
            url = "".join((url, resource))

        print('Downloading resource: {}'.format(url))

        path = os.path.join(destination_dir, resource)
        with urllib.request.urlopen(url) as response:
            if rsrc_atts['archive'] is not None:
                if rsrc_atts['archive'] == 'gz':
                    import tarfile
                    try:
                        tar = tarfile.open(mode='r:{}'.format(rsrc_atts['archive']), fileobj=response)
                    except IOError as e:
                        print(str(e))
                    else:
                        rsrcs = set(
                            self.model_atts.constituent_resource(con) for con in
                            self.model_atts.available_constituents()
                        )
                        tar.extractall(path=destination_dir, members=[m for m in tar.getmembers() if m.name in rsrcs])
                        tar.close()
                elif rsrc_atts['archive'] == 'zip':  # Unzip .zip files
                    zip_file = os.path.join(destination_dir, os.path.basename(resource) + '.zip')
                    with open(zip_file, 'wb') as out_file:
                        shutil.copyfileobj(response, out_file)
                    # Unzip the files
                    print("Unzipping files to: {}".format(destination_dir))
                    with ZipFile(zip_file, 'r') as unzipper:
                        # Extract all the files in the archive
                        unzipper.extractall(path=destination_dir)
                    print("Deleting zip file: {}".format(zip_file))
                    os.remove(zip_file)  # delete the zip file
            else:
                with open(path, 'wb') as f:
                    f.write(response.read())

        return path

    def download_model(self, resource_dir=None):
        """Download all of the model's resources for later use."""
        resources = set(
            self.model_atts.constituent_resource(con) for con in
            self.model_atts.available_constituents()
        )
        if not resource_dir:
            data_dir = self.model_atts.data_dir_name or self.model
            resource_dir = os.path.join(config['data_dir'], data_dir)
        for r in resources:
            path = os.path.join(resource_dir, r)
            if not os.path.exists(path):
                self.download(r, resource_dir)
        return resource_dir

    def remove_model(self):
        """Remove all of the model's resources."""
        data_dir = self.model_atts.data_dir_name or self.model
        resource_dir = os.path.join(config['data_dir'], data_dir)
        if os.path.exists(resource_dir):
            import shutil

            shutil.rmtree(resource_dir, ignore_errors=True)

    def get_datasets(self, constituents, filenames=None):
        """Returns a list of xarray datasets.

        Args:
            constituents (list[str]): List of the constiuent names to retrieve datasets for
            filenames (Optional[list[list[str]]]): Paths to the NetCDF files, parallel with return value if provided.
                Only needed by the FES2014 model currently.

        Returns:
            list[list[Dataset]]: The xarray Datasets for the requested constituents
        """
        available = self.available_constituents()
        if any(const not in available for const in constituents):
            raise ValueError('Constituent not recognized.')
        # handle compatible files together
        self.datasets = []
        for const_group in self.model_atts.constituent_groups():
            rsrcs = set(self.model_atts.constituent_resource(const) for const in set(constituents) & set(const_group))

            paths = set()
            if config['pre_existing_data_dir']:
                missing = set()
                data_dir = self.model_atts.data_dir_name or self.model
                for r in rsrcs:
                    path = os.path.join(config['pre_existing_data_dir'], data_dir, r)
                    paths.add(path) if os.path.exists(path) else missing.add(r)
                rsrcs = missing
                if not rsrcs and paths:
                    paths_list = list(paths)  # If the caller wants filenames, give them as parallel list with return.
                    self.datasets.append([xr.open_dataset(path) for path in paths_list])
                    if filenames is not None:
                        filenames.append(paths_list)
                    continue

            data_dir = self.model_atts.data_dir_name or self.model
            resource_dir = os.path.join(config['data_dir'], data_dir)
            for r in rsrcs:
                path = os.path.join(resource_dir, r)
                if not os.path.exists(path):
                    self.download(r, resource_dir)
                paths.add(path)

            if paths:
                paths_list = list(paths)
                self.datasets.append([xr.open_dataset(path) for path in paths_list])
                if filenames is not None:  # If the caller wants the filenames, give them as parallel list with return.
                    filenames.append(paths_list)

        return self.datasets



"""Initialize the module."""

# 1. Standard Python modules
import os

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules


config = {
    'pre_existing_data_dir': '',  # ignored if empty string
    # 'data_dir': os.path.join(os.path.dirname(__file__), 'data'),
    # If on Windows, use the system APPDATA directory to download resources to. The Python installation may
    # be in a protected folder. Default to the package directory if no APPDATA environment variable.
    'data_dir': os.path.join(os.getenv('APPDATA', os.path.dirname(os.path.dirname(__file__))), 'harmonica', 'data')
}

__version__ = '2.2.0'



"""Top-level interface for interacting with the tidal database models."""

# 1. Standard Python modules

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from .adcirc_database import AdcircDB
from .leprovost_database import LeProvostDB
from .resource import ResourceManager
from .tpxo_database import TpxoDB


class Constituents:
    """Class for extracting tidal constituent data.

    Attributes:
        _current_model (:obj:`tidal_database.TidalDB`): The tidal model currently being used for extraction

    """
    def __init__(self, model=ResourceManager.DEFAULT_RESOURCE):
        """Constructor tidal constituent extractor interface.

        Use this class as opposed to the lower level implementations

        Args:
            model (:obj:`str`, optional): Name of the tidal model. See resource.py for supported models.

        """
        self._current_model = None
        self.change_model(model)

    @property
    def data(self):
        """:obj:`Pandas.DataFrame` Access the underlying Pandas data frame of this constituent object."""
        if self._current_model:
            return self._current_model.data
        return []

    @data.setter
    def data(self, value):
        """Set the underlying dataframe of the current model."""
        if self._current_model:
            self._current_model.data = value

    def change_model(self, new_model):
        """Change the current tidal database model.

        Args:
            new_model (str): The model to switch to. See the constants defined in ResourceManager for valid values.
        """
        new_model = new_model.lower()
        if self._current_model and self._current_model.model == new_model:
            return  # Already have the correct impl and resources for this model, nothing to do.

        if new_model in ResourceManager.TPXO_MODELS:  # Switch to a TPXO model
            self._current_model = TpxoDB(new_model)
        elif new_model in ResourceManager.LEPROVOST_MODELS:
            self._current_model = LeProvostDB(new_model)
        elif new_model in ResourceManager.ADCIRC_MODELS:
            self._current_model = AdcircDB()
        else:
            tpxo_models = ", ".join(ResourceManager.TPXO_MODELS) + ", "
            leprovost_models = ", ".join(ResourceManager.LEPROVOST_MODELS) + ", "
            adcirc_models = ", ".join(ResourceManager.ADCIRC_MODELS)
            supported_models = tpxo_models + leprovost_models + adcirc_models
            raise ValueError(f'Model not supported: "{new_model}". Must be one of: {supported_models.strip()}.')

    def get_components(self, locs, cons=None, positive_ph=False, model=None):
        """Abstract method to get amplitude, phase, and speed of specified constituents at specified point locations.

        Args:
            locs (:obj:`list` of :obj:`tuple` of :obj:`float`): latitude [-90, 90] and longitude [-180 180] or [0 360]
                of the requested points.
            cons (:obj:`list` of :obj:`str`, optional): List of the constituent names to get amplitude and phase for. If
                not supplied, all valid constituents will be extracted.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).
            model (:obj:`str`, optional): Name of the tidal model to use to query for the data. If not provided, current
                model will be used. If a model other than the current is provided, current model is switched.

        Returns:
           :obj:`list` of :obj:`pandas.DataFrame`: Implementations should return a list of dataframes of constituent
                information including amplitude (meters), phase (degrees) and speed (degrees/hour, UTC/GMT). The list is
                parallel with locs, where each element in the return list is the constituent data for the corresponding
                element in locs. Empty list on error. Note that function uses fluent interface pattern.

        """
        if model and model.lower() != self._current_model.model:
            self.change_model(model.lower())
        return self._current_model.get_components(locs, cons, positive_ph)

    def get_nodal_factor(self, names, timestamp, timestamp_middle):
        """Get the nodal factor for specified constituents at a specified time.

        Args:
            names (:obj:`list` of :obj:`str`): Names of the constituents to get nodal factors for
            timestamp (:obj:`datetime.datetime`): Stat date and time to extract constituent arguments at
            timestamp_middle (:obj:`datetime.datetime`): Date and time to consider as the middle of the
                series.

        Returns:
            :obj:`pandas.DataFrame`: Constituent data frames. Each row contains frequency, earth tidal reduction factor,
                amplitude, nodal factor, and equilibrium argument for one of the specified constituents. Rows labeled by
                constituent name.

        """
        return self._current_model.get_nodal_factor(names, timestamp, timestamp_middle)



"""The reconstruct CLI command."""

# 1. Standard Python modules
import argparse
from datetime import date, datetime
import sys

# 2. Third party modules
import numpy as np
import pandas as pd
from pytides.tide import Tide as pyTide

# 3. Aquaveo modules

# 4. Local modules
from .common import add_common_args, add_const_out_args, add_loc_model_args
from ..harmonica import Tide


DESCR = 'Reconstruct the tides at specified location and times.'
EXAMPLE = """
Example:

    harmonica reconstruct 38.375789 -74.943915
"""


def validate_date(value):
    """Validate a date string.

    Args:
        value (str): The date string, should be in '%Y-%m-%d' format (e.g. '2022-02-20')

    Returns:
        bool: True if the value is positive, False if it is zero or negative
    """
    try:
        # return date.fromisoformat(value) # python 3.7
        return pd.datetime.strptime(value, '%Y-%m-%d')
    except ValueError:
        msg = "Not a valid date: '{0}'.".format(value)
        raise argparse.ArgumentTypeError(msg)


def check_positive(value):
    """Check if a value is positive.

    Args:
        value (Union[int, float]): The value to check

    Returns:
        bool: True if the value is positive, False if it is zero or negative
    """
    flt = float(value)
    if flt <= 0:
        msg = "Not a valid time length: {0}".format(value)
        raise argparse.ArgumentTypeError(msg)
    return flt


def config_parser(p, sub=False):
    """Configure the command line arguments passed the reconstruct CLI command.

    Args:
        p (ArgumentParser): The argument parser
        sub (Optional[bool]): True if this is a resources subparser
    """
    # Subparser info
    if sub:
        p = p.add_parser(
            'reconstruct',
            description=DESCR,
            help=DESCR,
            epilog=EXAMPLE,
            add_help=False,
        )

    add_common_args(p)
    p.add_argument(
        '-S', '--start_date',
        type=validate_date,
        default=date.today(),
        help='Start Date [YYYY-MM-DD], default: today'
    )
    p.add_argument(
        '-L', '--length',
        type=check_positive,
        default=7.,
        help='Length of series in days [positive non-zero], default: 7'
    )
    add_loc_model_args(p)
    add_const_out_args(p)


def parse_args(args):
    """Parse the command line arguments passed the reconstruct CLI command.

    Args:
        args (...): Variable length positional arguments

    Returns:
        ArgumentParser: The command line argument parser
    """
    p = argparse.ArgumentParser(
        description=DESCR,
        epilog=EXAMPLE,
        add_help=False,
    )
    config_parser(p)
    return p.parse_args(args)


def execute(args):
    """Execute the reconstruct CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    times = pyTide._times(datetime.fromordinal(args.start_date.toordinal()), np.arange(args.length * 24., dtype=float))
    tide = Tide(model=args.model).reconstruct_tide(loc=[args.lat, args.lon], times=times, cons=args.cons,
                                                   positive_ph=args.positive_phase)
    out = tide.data.to_csv(args.output, sep='\t', header=True, index=False)
    if args.output is None:
        print(out)
    print('\nComplete.\n')


def main(args=None):
    """Entry point for the reconstruct CLI command.

    Args:
        args (...): Variable length positional arguments
    """
    if not args:
        args = sys.argv[1:]
    try:
        execute(parse_args(args))
    except RuntimeError as e:
        print(str(e))
        sys.exit(1)
    return



"""Class for managing the ADCIRC 2015 tidal database model."""

# 1. Standard Python modules
import math

# 2. Third party modules
import numpy
import pandas as pd

# 3. Aquaveo modules
from xms.grid.geometry.tri_search import TriSearch

# 4. Local modules
from .resource import ResourceManager
from .tidal_database import convert_coords, get_complex_components, NOAA_SPEEDS, TidalDB


DEFAULT_ADCIRC_RESOURCE = 'adcirc2015'


class AdcircDB(TidalDB):
    """The class for extracting tidal data, specifically amplitude and phases, from an ADCIRC database."""
    def __init__(self, model=DEFAULT_ADCIRC_RESOURCE):
        """Constructor for the ADCIRC tidal database extractor.

        Args:
            model (:obj:`str`, optional): Name of the ADCIRC tidal database version. Currently defaults to the only
                supported release, 'adcirc2015'. Expand resource.adcirc_models for future versions.
        """
        model = model.lower()
        if model not in ResourceManager.ADCIRC_MODELS:
            raise ValueError("\'{}\' is not a supported ADCIRC model. Must be one of: {}.".format(
                model, ", ".join(ResourceManager.ADCIRC_MODELS).strip()
            ))
        super().__init__(model)

    def get_components(self, locs, cons=None, positive_ph=False):
        """Get the amplitude, phase, and speed for the given constituents at the given points.

        Args:
            locs (:obj:`list` of :obj:`tuple` of :obj:`float`): latitude [-90, 90] and longitude [-180 180] or [0 360]
                of the requested points.
            cons (:obj:`list` of :obj:`str`, optional): List of the constituent names to get amplitude and phase for. If
                not supplied, all valid constituents will be extracted.
            positive_ph (bool, optional): Indicate if the returned phase should be all positive [0 360] (True) or
                [-180 180] (False, the default).

        Returns:
            (:obj:`list` of :obj:`pandas.DataFrame`): A list of dataframes of constituent information including
                amplitude (meters), phase (degrees) and speed (degrees/hour, UTC/GMT). The list is parallel with locs,
                where each element in the return list is the constituent data for the corresponding element in locs.
                Note that function uses fluent interface pattern.
        """
        # pre-allocate the return value
        if not cons:
            cons = list(self.resources.available_constituents())  # Get all constituents by default
        else:
            cons = [con.upper() for con in cons]

        # Make sure point locations are valid lat/lon
        locs = convert_coords(locs)
        if not locs:
            return self  # ERROR: Not in latitude/longitude

        # Accumulate each point's constituent rows and build one DataFrame per point at the end, instead of
        # writing DataFrame cells one at a time.
        rows = [{} for _ in range(len(locs))]

        # Step 1: read the file and get geometry:
        con_dsets = self.resources.get_datasets(cons)[0]
        # ravel() collapses the trailing singleton dim (nodes stored as (n, 1)); tolist() gives plain Python floats
        # so building mesh_pts is a single vectorized read instead of millions of per-element numpy scalar casts.
        con_x = con_dsets[0].x.values.ravel()
        con_y = con_dsets[0].y.values.ravel()

        mesh_pts = [(x, y, 0.0) for x, y in zip(con_x.tolist(), con_y.tolist())]
        tri_list = con_dsets[0].element.values.flatten().tolist()
        tri_search = TriSearch(mesh_pts, tri_list)

        points_and_weights = []
        for i, pt in enumerate(locs):
            pt_flip = (pt[1], pt[0])
            tri_idx = tri_search.triangle_containing_point(pt_flip)
            if tri_idx != -1:
                pt_1 = int(tri_list[tri_idx])
                pt_2 = int(tri_list[tri_idx + 1])
                pt_3 = int(tri_list[tri_idx + 2])
                x1, y1, z1 = mesh_pts[pt_1]
                x2, y2, z2 = mesh_pts[pt_2]
                x3, y3, z3 = mesh_pts[pt_3]
                x = pt_flip[0]
                y = pt_flip[1]
                # Compute barocentric area weights
                ta = abs((x2 * y3 - x3 * y2) - (x1 * y3 - x3 * y1) + (x1 * y2 - x2 * y1))
                w1 = ((x - x3) * (y2 - y3) + (x2 - x3) * (y3 - y)) / ta
                w2 = ((x - x1) * (y3 - y1) - (y - y1) * (x3 - x1)) / ta
                w3 = ((y - y1) * (x2 - x1) - (x - x1) * (y2 - y1)) / ta
                points_and_weights.append((i, (pt_1, pt_2, pt_3), (w1, w2, w3)))
            else:  # Outside domain, return NaN for all constituents
                for con in cons:
                    rows[i][con] = (numpy.nan, numpy.nan, numpy.nan)

        for con in cons:
            con_amp_name = con + "_amplitude"
            con_pha_name = con + "_phase"
            # Materialize the per-node arrays once per constituent; the point loop then indexes numpy directly
            # instead of paying xarray scalar-access overhead for every corner of every point.
            con_amp = con_dsets[0][con_amp_name].values
            con_pha = con_dsets[0][con_pha_name].values
            for i, pts, weights in points_and_weights:
                amps = [float(con_amp[pts[0]]), float(con_amp[pts[1]]), float(con_amp[pts[2]])]
                phases = [
                    math.radians(float(con_pha[pts[0]])),
                    math.radians(float(con_pha[pts[1]])),
                    math.radians(float(con_pha[pts[2]])),
                ]

                # Get the real and imaginary components from the amplitude and phases in the file. It
                # would be better if these values were stored in the file like TPXO.
                components = get_complex_components(amps, phases)

                # Perform area weighted interpolation
                ctr = components[0][0] * weights[0] + components[1][0] * weights[1] + components[2][0] * weights[2]
                cti = components[0][1] * weights[0] + components[1][1] * weights[1] + components[2][1] * weights[2]
                new_amp = math.sqrt(ctr * ctr + cti * cti)

                # Compute interpolated phase
                if new_amp == 0.0:
                    new_phase = 0.0
                else:
                    new_phase = math.degrees(math.acos(ctr / new_amp))
                    if cti < 0.0:
                        new_phase = 360.0 - new_phase
                speed = NOAA_SPEEDS[con][0] if con in NOAA_SPEEDS else numpy.nan
                rows[i][con] = (new_amp, new_phase, speed)

        self.data = [
            pd.DataFrame.from_dict(row, orient='index', columns=['amplitude', 'phase', 'speed'])
            for row in rows
        ]
        return self



"""The constituents CLI command."""

# 1. Standard Python modules
import argparse
import sys

# 2. Third party modules

# 3. Aquaveo modules

# 4. Local modules
from .common import add_common_args
from .main_constituents import config_parser as config_parser_constituents
from .main_deconstruct import config_parser as config_parser_deconstruct
from .main_reconstruct import config_parser as config_parser_reconstruct
from .main_resources import config_parser as config_parser_resources


def main():
    """Entry point for the top-level CLI command.

    Returns:
        int: 0 on success, 1 on error
    """
    p = argparse.ArgumentParser(
        description='harmonica is a tool for working with tidal harmonics.',
        add_help=False,
    )
    add_common_args(p)
    sps = p.add_subparsers(
        metavar='command',
        dest='cmd',
    )
    sps.required = True
    config_parser_constituents(sps, True)
    config_parser_deconstruct(sps, True)
    config_parser_reconstruct(sps, True)
    config_parser_resources(sps, True)

    args = p.parse_args(sys.argv[1:])
    try:
        sys.modules[f'harmonica.cli.main_{args.cmd}'].execute(args)
    except RuntimeError as e:
        print(str(e))
        sys.exit(1)
    return 0


if __name__ == '__main__':
    sys.exit(main())



"""Build or install the harmonica package."""
# 1. Standard Python modules
import os

# 2. Third party modules
from setuptools import setup

# 3. Aquaveo modules

# 4. Local modules


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


install_requires = [
    'dask',
    'netCDF4',
    'numpy',
    'pandas',
    'pytides>=1.0.0',  # ERDC fork packaged by Aquaveo
    'toolz',
    'xarray',
    'xmsgrid>=6.0.0',
]

extras_require = {
    'build': [
        'setuptools',
    ],
    'tests': [],
}

extras_require['all'] = sorted(set(sum(extras_require.values(), [])))

entry_points = [
    'harmonica = harmonica.cli.main:main',
    'harmonica-constituents = harmonica.cli.main_constituents:main',
    'harmonica-deconstruct = harmonica.cli.main_deconstruct:main',
    'harmonica-reconstruct = harmonica.cli.main_reconstruct:main',
    'harmonica-resources = harmonica.cli.main_resources:main',
]


version = '2.2.0'


setup(
    name='harmonica',
    version=version,
    description='Worldwide amplitude, phase, and speed for standard tidal constituents and tidal time series '
                'reconstruction and deconstruction.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Kevin Winters',
    author_email='Kevin.D.Winters@erdc.dren.mil',
    maintainer='Aquaveo LLC',
    url='https://github.com/aquaveo/harmonica',
    packages=['harmonica', 'harmonica.cli'],
    dependency_links=[
        'https://public.aquapi.aquaveo.com/aquaveo/stable'
        'https://public.aquapi.aquaveo.com/aquaveo/stable/pytides',
        'https://public.aquapi.aquaveo.com/aquaveo/stable/xmsgrid',
    ],
    entry_points={
        'console_scripts': entry_points
    },
    install_requires=install_requires,
    extras_require=extras_require,
    tests_require=extras_require['tests'],
    test_suite='tests',
    keywords='harmonica',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering'
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires='>=3.10'
)
